#!/usr/bin/env python3

"""
@author Irumi Sugimori
@author Shuji Kosuge
Time-stamp: <2023-06-23 12:42:29 irumisugimori>
"""

import sys
import getopt
import re
import os
from util_col2csp import *


def _usage(msg=""):
    if msg != "":
        print(msg)
        print("")
    print(f"Usage: python3 {sys.argv[0]} <options> col_file")
    print(f"    -h --help               : show this help")
    print(f"    -m --model <number>     : select model (default: {config.model})")
    exit()


class Config:
    def __init__(self):
        self.model = 1


config = Config()


"""
solvers
"""

def solver_one_pair_basic(nodes, edges, start, goal, max_length):
    print_comment(f"{config.model}:", "one pair basic modeling")
    edges, start, goal = initialization_one_pair_basic(edges, start, goal)
    edge_variables_one_pair_basic(edges)
    node_label_variables_one_pair_basic(nodes, start, goal, max_length)
    indegree_constraints_of_the_start_node_one_pair_basic(edges, start)
    outdegree_constraints_of_the_start_node_one_pair_basic(edges, start)
    indegree_constraints_of_the_goal_node_one_pair_basic(edges, goal)
    outdegree_constraints_of_the_goal_node_one_pair_basic(edges, goal)
    indegree_constraints_of_nonterminal_nodes_one_pair_basic(nodes, edges, start, goal)
    outdegree_constraints_of_nonterminal_nodes_one_pair_basic(nodes, edges, start, goal)
    degree_constraints_of_nonterminal_nodes_one_pair_basic(nodes, edges, start, goal)
    connectiviy_constraints_one_pair_basic(edges)
    path_length_constraints_one_pair_basic(edges, max_length)
    
    
def solver_one_pair_improved(nodes, edges, start, goal, max_length):
    print_comment(f"{config.model}:", "one pair improved modeling")
    nodes, edges, start, goal = initialization_one_pair_improved(nodes, edges, start, goal, max_length)
    edge_variables_one_pair_basic(edges)
    node_label_variables_one_pair_improved(nodes, start, goal, max_length)
    node_variables_one_pair_improved(nodes)
    channeling_constraints_between_node_variables_and_edge_variables_one_pair_improved(edges)
    channeling_constraints_between_node_variables_and_node_label_variables_one_pair_improved(nodes, start, goal)
    indegree_constraints_of_the_start_node_one_pair_basic(edges, start)
    outdegree_constraints_of_the_start_node_one_pair_basic(edges, start)
    indegree_constraints_of_the_goal_node_one_pair_basic(edges, goal)
    outdegree_constraints_of_the_goal_node_one_pair_basic(edges, goal)
    indegree_constraints_of_nonterminal_nodes_one_pair_basic(nodes, edges, start, goal)
    outdegree_constraints_of_nonterminal_nodes_one_pair_basic(nodes, edges, start, goal)
    degree_constraints_of_nonterminal_nodes_one_pair_basic(nodes, edges, start, goal)
    connectiviy_constraints_one_pair_basic(edges)
    path_length_constraints_one_pair_basic(edges, max_length)
    in_path_node_constraints_one_pair_improved(nodes, max_length)
    forbidden_node_pair_constraints_one_pair_improved(nodes, start, goal, max_length)
    
    
def solver_all_pairs_basic(nodes, edges, max_length):
    print_comment(f"{config.model}:", "all pairs basic modeling")
    edges = initialization_all_pairs_basic(edges)
    start_node_variables_all_pairs_basic(nodes)
    goal_node_variables_all_pairs_basic(nodes)
    terminal_constraints_all_pairs_basic()
    edge_variables_all_pairs_basic(edges)
    node_label_variables_all_pairs_basic(nodes, max_length)
    indegree_constraints_all_pairs_basic(nodes, edges)
    outdegree_constraints_all_pairs_basic(nodes, edges)
    indegree_constraints_of_the_start_node_all_pairs_basic(nodes, edges)
    outdegree_constraints_of_the_start_node_all_pairs_basic(nodes, edges)
    indegree_constraints_of_the_goal_node_all_pairs_basic(nodes, edges)
    outdegree_constraints_of_the_goal_node_all_pairs_basic(nodes, edges)
    degree_constraints_of_nonterminal_nodes_all_pairs_basic(nodes, edges)
    connectiviy_constraints_all_pairs_basic(edges)
    path_length_constraints_all_pairs_basic(edges, max_length)


def solver_all_pairs_improved_without_38(nodes, edges, max_length):
    print_comment(f"{config.model}:", "all pairs improved modeling without (38)")
    edges = initialization_all_pairs_improved(nodes, edges)
    start_node_variables_all_pairs_basic(nodes)
    goal_node_variables_all_pairs_basic(nodes)
    terminal_constraints_all_pairs_basic()
    terminal_combination_constraints_all_pairs_improved(nodes, edges, max_length)
    edge_variables_all_pairs_basic(edges)
    node_label_variables_all_pairs_basic(nodes, max_length)
    node_variables_all_pairs_improved(nodes)
    channeling_constraints_between_node_variables_and_edge_variables_all_pairs_improved(edges)
    channeling_constraints_between_node_variables_and_node_label_variables_all_pairs_improved(nodes)
    indegree_constraints_all_pairs_basic(nodes, edges)
    outdegree_constraints_all_pairs_basic(nodes, edges)
    indegree_constraints_of_the_start_node_all_pairs_basic(nodes, edges)
    outdegree_constraints_of_the_start_node_all_pairs_basic(nodes, edges)
    indegree_constraints_of_the_goal_node_all_pairs_basic(nodes, edges)
    outdegree_constraints_of_the_goal_node_all_pairs_basic(nodes, edges)
    degree_constraints_of_nonterminal_nodes_all_pairs_basic(nodes, edges)
    connectiviy_constraints_all_pairs_basic(edges)
    path_length_constraints_all_pairs_basic(edges, max_length)
    in_path_node_constraints_all_pairs_improved(nodes, max_length)
    forbidden_node_pair_constraints_all_pairs_improved(nodes, max_length)

    
def solver_all_pairs_improved_with_38(nodes, edges, max_length):
    print_comment(f"{config.model}:", "all pairs improved modeling with (38)")
    edges = initialization_all_pairs_improved(nodes, edges)
    start_node_variables_all_pairs_basic(nodes)
    goal_node_variables_all_pairs_basic(nodes)
    terminal_constraints_all_pairs_basic()
    terminal_combination_constraints_all_pairs_improved(nodes, edges, max_length)
    edge_variables_all_pairs_basic(edges)
    node_label_variables_all_pairs_basic(nodes, max_length)
    node_variables_all_pairs_improved(nodes)
    channeling_constraints_between_node_variables_and_edge_variables_all_pairs_improved(edges)
    channeling_constraints_between_node_variables_and_node_label_variables_all_pairs_improved(nodes)
    indegree_constraints_all_pairs_basic(nodes, edges)
    outdegree_constraints_all_pairs_basic(nodes, edges)
    indegree_constraints_of_the_start_node_all_pairs_basic(nodes, edges)
    outdegree_constraints_of_the_start_node_all_pairs_basic(nodes, edges)
    indegree_constraints_of_the_goal_node_all_pairs_basic(nodes, edges)
    outdegree_constraints_of_the_goal_node_all_pairs_basic(nodes, edges)
    degree_constraints_of_nonterminal_nodes_all_pairs_basic(nodes, edges)
    connectiviy_constraints_all_pairs_basic(edges)
    path_length_constraints_all_pairs_basic(edges, max_length)
    in_path_node_constraints_all_pairs_improved(nodes, max_length)
    forbidden_node_pair_constraints_all_pairs_improved(nodes, max_length)
    forbidden_edge_constraints_all_pairs_improved(nodes, edges, max_length)


"""
main
"""

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hm:", ["help", "model="])
        for opt, arg in opts:
            if opt in ("-h", "--help"):
                _usage()
            elif opt in ("-m", "--model"):
                config.model = int(arg)
        if len(args) != 1:
            _usage("please input col-file.")

    except getopt.GetoptError as err:
        _usage(str(err))

    if config.model not in [1,2,3]:
        raise ValueError("ERROR! Model value should range from 1 to 3.")

    col_file = args[0]
    nodes = []
    edges = []
    start = -1
    goal = -1
    max_length = -1
    with open(col_file) as cf:
        for line in cf:
            p = re.match(r"^p\sedge\s(\d+)\s(\d+)", line)
            if p is not None:
                nodes = [i for i in range(1,int(p.group(1))+1)]
                continue
            e = re.match(r"^e\s(\d+)\s(\d+)", line)
            if e is not None:
                edges.append((int(e.group(1)), int(e.group(2))))
                continue
            l = re.match(r"^l\s(\d+)", line)
            if l is not None:
                max_length = int(l.group(1))
                continue
            t = re.match(r"^t\s(\d+)\s(\d+)", line)
            if t is not None:
                start = int(t.group(1))
                goal = int(t.group(2))
                continue
            print_comment(line[:-1])

    print_comment(os.path.basename(col_file))

    if start > 0:
        if config.model == 1:
            solver_one_pair_basic(nodes, edges, start, goal, max_length)
        elif config.model in [2,3]:
            solver_one_pair_improved(nodes, edges, start, goal, max_length)
    else:
        if config.model == 1:
            solver_all_pairs_basic(nodes, edges, max_length)
        elif config.model == 2:
            solver_all_pairs_improved_without_38(nodes, edges, max_length)
        elif config.model == 3:
            solver_all_pairs_improved_with_38(nodes, edges, max_length)

    print(";")
    print_comment("END")

    
if __name__ == "__main__":
    main()
